@@ -113,12 +113,13 @@ func binaryBigFloatOp(c Context, u Value, op func(*big.Float, *big.Float, *big.F
 // Also we need a context (really a config); see the bigIntExpOp function below.
 // We know this is not 0**negative.
 func bigIntExp(c Context, i, j, k *big.Int) *big.Int {
-	if j.Cmp(bigOne.Int) == 0 || j.Sign() == 0 {
+	if j.Cmp(bigIntOne.Int) == 0 || j.Sign() == 0 {
 		return i.Set(j)
 	}
 	// -1ⁿ is just parity.
-	if j.Cmp(bigMinusOne.Int) == 0 {
-		if k.And(k, bigOne.Int).Int64() == 0 {
+	if j.Cmp(bigIntMinusOne.Int) == 0 {
+		var x big.Int
+		if x.And(k, bigIntOne.Int).Int64() == 0 {
 			return i.Neg(j)
 		}
 		return i.Set(j)
@@ -133,7 +134,7 @@ func bigIntExp(c Context, i, j, k *big.Int) *big.Int {
 		exp = -exp
 	}
 	// "2" is just shift. math/big should do this, really.
-	if j.Cmp(bigTwo.Int) == 0 && exp >= 0 {
+	if j.Cmp(bigIntTwo.Int) == 0 && exp >= 0 {
 		return i.Lsh(big.NewInt(1), uint(exp))
 	}
 	mustFit(c.Config(), int64(j.BitLen())*exp)
@@ -169,6 +170,8 @@ func toBool(t Value) bool {
 		return t.Sign() != 0
 	case BigFloat:
 		return t.Sign() != 0
+	case Complex:
+		return !isZero(t.real) || !isZero(t.imag)
 	}
 	Errorf("cannot convert %T to bool", t)
 	panic("not reached")
@@ -185,23 +188,33 @@ func andBool(t Value) bool {
 		}
 		return true
 	}
-	return t.(Int) == Int(1)
+	return t.(Int) == one
 }
 
-var (
-	zero        = Int(0)
-	one         = Int(1)
-	minusOne    = Int(-1)
-	bigZero     = bigInt64(0)
-	bigOne      = bigInt64(1)
-	bigTwo      = bigInt64(2)
-	bigMinusOne = bigInt64(-1)
-)
-
 var BinaryOps = make(map[string]BinaryOp)
 
 func init() {
 	var ops = []*binaryOp{
+		{
+			name:        "j",
+			elementwise: true,
+			whichType:   binaryArithType,
+			fn: [numType]binaryFn{
+				intType: func(c Context, u, v Value) Value {
+					return newComplex(u, v)
+				},
+				bigIntType: func(c Context, u, v Value) Value {
+					return newComplex(u, v)
+				},
+				bigRatType: func(c Context, u, v Value) Value {
+					return newComplex(u, v)
+				},
+				bigFloatType: func(c Context, u, v Value) Value {
+					return newComplex(u, v)
+				},
+			},
+		},
+
 		{
 			name:        "+",
 			elementwise: true,
@@ -221,6 +234,9 @@ func init() {
 				bigFloatType: func(c Context, u, v Value) Value {
 					return binaryBigFloatOp(c, u, (*big.Float).Add, v)
 				},
+				complexType: func(c Context, u, v Value) Value {
+					return u.(Complex).add(c, v.(Complex))
+				},
 			},
 		},
 
@@ -243,6 +259,9 @@ func init() {
 				bigFloatType: func(c Context, u, v Value) Value {
 					return binaryBigFloatOp(c, u, (*big.Float).Sub, v)
 				},
+				complexType: func(c Context, u, v Value) Value {
+					return u.(Complex).sub(c, v.(Complex))
+				},
 			},
 		},
 
@@ -264,6 +283,9 @@ func init() {
 				bigFloatType: func(c Context, u, v Value) Value {
 					return binaryBigFloatOp(c, u, (*big.Float).Mul, v)
 				},
+				complexType: func(c Context, u, v Value) Value {
+					return u.(Complex).mul(c, v.(Complex))
+				},
 			},
 		},
 
@@ -281,6 +303,9 @@ func init() {
 				bigFloatType: func(c Context, u, v Value) Value {
 					return binaryBigFloatOp(c, u, (*big.Float).Quo, v)
 				},
+				complexType: func(c Context, u, v Value) Value {
+					return u.(Complex).div(c, v.(Complex))
+				},
 			},
 		},
 
@@ -301,8 +326,9 @@ func init() {
 					}
 					return binaryBigIntOp(u, (*big.Int).Quo, v) // Go-like division.
 				},
-				bigRatType:   nil, // Not defined for rationals. Use div.
+				bigRatType:   nil, // Not defined for rationals.
 				bigFloatType: nil,
+				complexType:  nil,
 			},
 		},
 
@@ -325,6 +351,7 @@ func init() {
 				},
 				bigRatType:   nil, // Not defined for rationals. Use mod.
 				bigFloatType: nil,
+				complexType:  nil,
 			},
 		},
 
@@ -341,6 +368,7 @@ func init() {
 				},
 				bigRatType:   nil, // Not defined for rationals. Use div.
 				bigFloatType: nil,
+				complexType:  nil,
 			},
 		},
 
@@ -357,6 +385,7 @@ func init() {
 				},
 				bigRatType:   nil, // Not defined for rationals. Use mod.
 				bigFloatType: nil,
+				complexType:  nil,
 			},
 		},
 
@@ -373,11 +402,14 @@ func init() {
 						if u.(BigInt).Sign() == 0 {
 							Errorf("negative exponent of zero")
 						}
-						v = c.EvalUnary("abs", v).toType("**", c.Config(), bigIntType)
+						if isNegative(v) {
+							// Need the absolute value.
+							v = BigInt{big.NewInt(0).Neg(v.(BigInt).Int)}
+						}
 						return c.EvalUnary("/", binaryBigIntOp(u, bigIntExpOp(c), v))
 					}
 					x := u.(BigInt).Int
-					if x.Cmp(bigOne.Int) == 0 || x.Sign() == 0 {
+					if x.Cmp(bigIntOne.Int) == 0 || x.Sign() == 0 {
 						return u
 					}
 					return binaryBigIntOp(u, bigIntExpOp(c), v)
@@ -415,6 +447,7 @@ func init() {
 					return z.shrink()
 				},
 				bigFloatType: func(c Context, u, v Value) Value { return power(c, u, v) },
+				complexType:  func(c Context, u, v Value) Value { return power(c, u, v) },
 			},
 		},
 
@@ -427,6 +460,7 @@ func init() {
 				bigIntType:   logBaseU,
 				bigRatType:   logBaseU,
 				bigFloatType: logBaseU,
+				complexType:  logBaseU,
 			},
 		},
 
@@ -439,10 +473,10 @@ func init() {
 					a := int64(u.(Int))
 					b := int64(v.(Int))
 					if a == 0 || b == 0 || a == b {
-						return bigOne
+						return one
 					}
 					if a < 0 || b < 0 || a > b {
-						return bigZero
+						return zero
 					}
 					aFac := factorial(a)
 					bFac := factorial(b)
@@ -549,6 +583,13 @@ func init() {
 					i, j := u.(BigFloat), v.(BigFloat)
 					return toInt(i.Cmp(j.Float) == 0)
 				},
+				complexType: func(c Context, u, v Value) Value {
+					i, j := u.(Complex), v.(Complex)
+					if c.EvalBinary(i.real, "==", j.real) == Int(0) {
+						return Int(0)
+					}
+					return c.EvalBinary(i.imag, "==", j.imag)
+				},
 			},
 		},
 
@@ -575,6 +616,13 @@ func init() {
 					i, j := u.(BigFloat), v.(BigFloat)
 					return toInt(i.Cmp(j.Float) != 0)
 				},
+				complexType: func(c Context, u, v Value) Value {
+					i, j := u.(Complex), v.(Complex)
+					if c.EvalBinary(i.real, "!=", j.real) == Int(1) {
+						return Int(1)
+					}
+					return c.EvalBinary(i.imag, "!=", j.imag)
+				},
 			},
 		},
 
@@ -702,6 +750,9 @@ func init() {
 				bigFloatType: func(c Context, u, v Value) Value {
 					return toInt(toBool(u) && toBool(v))
 				},
+				complexType: func(c Context, u, v Value) Value {
+					return toInt(toBool(u) && toBool(v))
+				},
 			},
 		},
 
@@ -725,6 +776,9 @@ func init() {
 				bigFloatType: func(c Context, u, v Value) Value {
 					return toInt(toBool(u) || toBool(v))
 				},
+				complexType: func(c Context, u, v Value) Value {
+					return toInt(toBool(u) || toBool(v))
+				},
 			},
 		},
 
@@ -748,6 +802,9 @@ func init() {
 				bigFloatType: func(c Context, u, v Value) Value {
 					return toInt(toBool(u) != toBool(v))
 				},
+				complexType: func(c Context, u, v Value) Value {
+					return toInt(toBool(u) != toBool(v))
+				},
 			},
 		},
 
@@ -771,6 +828,9 @@ func init() {
 				bigFloatType: func(c Context, u, v Value) Value {
 					return toInt(!(toBool(u) && toBool(v)))
 				},
+				complexType: func(c Context, u, v Value) Value {
+					return toInt(!(toBool(u) && toBool(v)))
+				},
 			},
 		},
 
@@ -794,6 +854,9 @@ func init() {
 				bigFloatType: func(c Context, u, v Value) Value {
 					return toInt(!(toBool(u) || toBool(v)))
 				},
+				complexType: func(c Context, u, v Value) Value {
+					return toInt(!(toBool(u) || toBool(v)))
+				},
 			},
 		},
 
@@ -910,22 +973,11 @@ func init() {
 					// 1 0 1
 					// If they are negative the answers disagree with APL because
 					// of how modulo arithmetic works.
-					mod := func(b, a Value) Value {
-						if z, ok := a.(Int); ok && z == 0 {
-							return b
-						}
-						return c.EvalBinary(b, "mod", a)
-					}
-					div := func(b, a Value) Value {
-						if z, ok := a.(Int); ok && z == 0 {
-							return b
-						}
-						return c.EvalBinary(b, "div", a)
-					}
+					const op = "encode"
 					A, B := u.(Vector), v.(Vector)
 					// Scalar.
 					if len(A) == 1 && len(B) == 1 {
-						return mod(B[0], A[0])
+						return emod(op, c, B[0], A[0])
 					}
 					// Vector.
 					if len(B) == 1 {
@@ -934,8 +986,8 @@ func init() {
 						b := B[0]
 						for i := len(A) - 1; i >= 0; i-- {
 							a := A[i]
-							elems[i] = mod(b, a)
-							b = div(b, a)
+							elems[i] = emod(op, c, b, a)
+							b = ediv(op, c, b, a)
 						}
 						return NewVector(elems)
 					}
@@ -944,7 +996,7 @@ func init() {
 						elems := make([]Value, len(B))
 						a := A[0]
 						for i := range B {
-							elems[i] = mod(B[i], a)
+							elems[i] = emod(op, c, B[i], a)
 						}
 						return NewVector(elems)
 					}
@@ -959,36 +1011,25 @@ func init() {
 							b := B[j]
 							for i := len(A) - 1; i >= 0; i-- {
 								a := A[i]
-								elems[j+i*len(B)] = mod(b, a)
-								b = div(b, a)
+								elems[j+i*len(B)] = emod(op, c, b, a)
+								b = ediv(op, c, b, a)
 							}
 						}
 					})
 					return NewMatrix(shape, elems)
 				},
 				matrixType: func(c Context, u, v Value) Value {
-					mod := func(b, a Value) Value {
-						if z, ok := a.(Int); ok && z == 0 {
-							return b
-						}
-						return c.EvalBinary(b, "mod", a)
-					}
-					div := func(b, a Value) Value {
-						if z, ok := a.(Int); ok && z == 0 {
-							return b
-						}
-						return c.EvalBinary(b, "div", a)
-					}
 					A, B := u.(Vector), v.(*Matrix)
 					elems := make([]Value, len(A)*len(B.data))
 					shape := append([]int{len(A)}, B.Shape()...)
+					const op = "encode"
 					pfor(true, len(A), len(B.data), func(lo, hi int) {
 						for j := lo; j < hi; j++ {
 							b := B.data[j]
 							for i := len(A) - 1; i >= 0; i-- {
 								a := A[i]
-								elems[j+i*len(B.data)] = mod(b, a)
-								b = div(b, a)
+								elems[j+i*len(B.data)] = emod(op, c, b, a)
+								b = ediv(op, c, b, a)
 							}
 						}
 					})
